package jp.cqpub.transistor;

import java.io.IOException;
import java.net.*;

public class Udp
{
	//	Common variables
	private DatagramSocket socket;	// Socket work
	private DatagramPacket receivePacket;	// Receive packet object
	private DatagramPacket sendPacket;	// Send packet object
	private InetSocketAddress DestAddress;	// Destination address
	private byte[] recbuf, sendbuf;	// Data buffer
	private byte[][] rec_ring;	// Receive ring buffer
	private int[] rec_port;		// Port number
	private InetAddress[] rec_ip;	// IP address
	private int[] rec_length;	// Data length
	private volatile int queuein, queueout;	// Queue pointer
	private Thread net_thread;	// Thread object
	private volatile boolean abortflag;	// Thread terminate flag
	//
	//	Constructor
	//
	public Udp(String ad, int destport, int srcport)
	{
		try
		{	// Create send data object
			socket = new DatagramSocket(srcport);	// Specify sender port
		}
		catch (SocketException e) {return;}	// exit if fail
		DestAddress = new InetSocketAddress(ad, destport);
  	recbuf = new byte[1500];	// Receive buffer
  	sendbuf = new byte[1500];	// Send buffer
  	rec_ring = new byte[10][1500];	// Receive ring buffer
  	rec_port = new int[10];	// Sender port buffer
  	rec_ip = new InetAddress[10];	// Sender port buffer
  	rec_length = new int[10];	// Length buffer
	  receivePacket = new DatagramPacket(recbuf, recbuf.length);	// Create receive object
	  try
		{	// Create send object
			sendPacket = new DatagramPacket(sendbuf, sendbuf.length, DestAddress);
		}
		catch (SocketException e1) {	}	// Ignore error
		abortflag = false;	// Clear abort flag before thread start
		net_thread = new Thread(new Runnable()	// Create network thread
		{
			@Override
			public void run()	// Auto run object
			{
				for (;!abortflag;)	// Repeat forever
				{
					if(Receive() == false)	// Try to receive network packet (will be blocked)
						break;
					try
					{
						Thread.sleep(10);	// Sleep
					}
					catch (InterruptedException e){	}// Ignore error(never happens)
				}
				abortflag = false;
			}
		});
		net_thread.start();	// Start network thread
	}
	//
	//	Send packet
	//	
	public boolean Send(byte[] data, int len)
	{
		if(sendbuf == null)	// Valid buffer specified?
			return false;	// No then do nothing
		System.arraycopy(data, 0, sendbuf, 0, len);// Make copy of send data
		sendPacket.setLength(len);	// Set data length to send
		try
		{
			socket.send(sendPacket);	// Send data
		}
		catch (SocketException e)	{return false;}	// Error return
		catch (IOException e)	{return false;}	// Error return
		return true;	// Normal return
	}
	//
	//	Receive data and copy to receive queue
	//
	public boolean Receive()
	{
		int queuecou;
		if(socket == null)	// fail to create socket?
			return false;	// Do nothing
		try
		{
			socket.receive(receivePacket);	// Wait until network packet has come
			queuecou = queuein - queueout;	// Get queue available space
			if(queuecou < 0)	// If negative number?
				queuecou += 10;	// Make it plus
			if(queuecou < 9)	// Have space to put?
			{
				System.arraycopy(recbuf, 0, rec_ring[queuein], 0, recbuf.length);	// Save data
				rec_length[queuein] = receivePacket.getLength();	// Save data length
				rec_ip[queuein] = receivePacket.getAddress();	// Save source address
				rec_port[queuein] = receivePacket.getPort();	// Save port number
				if(queuein == 9)	// Increment queue in
					queuein = 0;	// Wrap round
				else
					queuein++;	// Just advance queue entry pointer
			}
		}
		catch (SocketException e) {return false;}// Socket error return
		catch (IOException e) {return false;}		// I/O error return
		return true;	// Normal return
	}
	//
	//	Get queued count
	//
	public int GetQueueCount()
	{
		int queuecou = queuein - queueout;	// Get queue count
		if(queuecou < 0)	// Negative?
			queuecou += 10;	// Make it plus
		return queuecou;	// Return with queued bynber
	}
	//
	//	Get queued data
	//
	public boolean GetData(byte[] result)
	{
		if(queuein != queueout)	// Check queue has data
		{	// Copy queued data
			System.arraycopy(rec_ring[queueout], 0, result, 0, rec_length[queueout]);
			if(queueout == 9)	// Increment queue pointer
				queueout = 0;	// Wrap round
			else
				queueout++;	// Just increment queue pointer
			return true;	// Good result
		}
		return false;	// No data has arrived yet
	}
	//
	//	Get sender port
	//
	public int GetLength()
	{
		if(queuein != queueout)	// Queue has data?
			return rec_length[queueout];	// Return with received data length	
		else
			return 0;	// return 0 if no data has arrived
	}
	//
	//	Get sender port
	//
	public int GetPort()
	{
		if(queuein != queueout)	// Queue has data?
			return rec_port[queueout];	// Get sender port number
		else
			return 0;	// return zero if no data has come
	}
	//
	//	Get sender ip address
	//
	public InetAddress GetAddress()
	{
		if(queuein != queueout)		// Queue has data?
			return rec_ip[queueout];	// Get sender IP address number & return
		else
			return null;	// null result if no data is received
	}
	//
	//	Abort receive operation	
	//	
	public void Abort()
	{
		abortflag = true;	// Set abort flag
		if(socket != null)	// Socket exists?	
			socket.close();	// Close socket and force stop receive
		while(abortflag);	// Confirm thread finished successfully
	}
}
